# rechter/bots/__init__.py
#
#

""" bots unit. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from rechter.utils.parse import txt_parse
from rechter.utils.time import short_date
from rechter.dispatcher import Dispatcher
from rechter.main import main, fleet
from rechter.defines import BLA, ENDC
from rechter.object import Object
from rechter.errors import error

from rechter.errors import RemoteDisconnect, NotImplemented

import logging
import socket
import time

## BOTS

class Bot(Dispatcher):

    """ Basic Bot. """

    default = ""
    cc = ""

    def __init__(zelf, *args, **kwargs):
        Dispatcher.__init__(zelf, *args, **kwargs)
        zelf._start = time.time()
        zelf.channels = ["#rechter", ]

    ## OUTPUT

    def _raw(zelf, *args, **kwargs): print(args[0])

    def say(zelf, *args, **kwargs): zelf._raw(args[1])
        
    def announce(zelf, *args, **kwargs):
        """ announce on channels. """
        zelf._last = time.time()
        logging.info("# %s/announce %s" % (zelf.type, " ".join(zelf.channels)))
        for channel in zelf.channels: zelf.say(channel, args[0])

    # CHANNELS

    def join(zelf, *args, **kwargs): pass
        
    def joining(zelf, *args, **kwargs):
        """ join channels.  """ 
        for channel in zelf.channels: zelf.join(channel)

    ## LOOP

    def start(zelf, *args, **kwargs):
        logging.warn("# %s/start" % zelf.type)
        zelf._status = "ok"
        if "collect" in kwargs: logging.warn("# %s/enable collect" % zelf.type)
        while zelf._status:
            try:
                event = zelf.event()
                logging.debug("! event %s" % event.json())
                logging.debug("! target %s" % event._target)
                if not zelf._status: break
                if not event: continue
                if "txt" in event: event.parsed = txt_parse(zelf, event.txt)
                else: event.parsed = txt_parse(zelf, "")
                thrs = zelf.dispatch(event)
                if thrs and kwargs.get("collect", False):
                    result = zelf.collect(thrs)
                    if result: logging.warn("# results %s " % ", ".join([str(x) for x in result]))
            except RemoteDisconnect: time.sleep(5.0) ; continue
            except (EOFError, KeyboardInterrupt): raise
            except: error()
        zelf.ready()
        logging.warn("# stop %s/%s" % zelf.type, event.cmnd)

    def event(zelf, *args, **kwargs):
        event = Object()
        event._target = zelf
        event.origin = "root@zelf"
        event.txt = args[0]
        return event

    ## DISPATCH

    def single(zelf, *args, **kwargs):
        if not args[0]: return
        event = Object()
        event._target = zelf
        event.origin = "rechter@shell"
        event.txt = args[0]
        event.parsed = txt_parse(zelf, event.txt)
        return zelf.dispatch(event)

    def dispatch(zelf, *args, **kwargs):
        event = args[0]
        if "txt" not in event: return
        if "parsed" not in event: event.parsed = txt_parse(zelf, event.txt)
        cmnds = zelf.find_cmnd(event.parsed.cmnd)
        thrs = []
        for cmnd in cmnds:
            try: o = zelf.put(cmnd.func, event) ; thrs.append(o)
            except AttributeError: continue
            except: error()
        return thrs

    def find_cmnd(zelf, *args, **kwargs):
        from rechter.main import main
        cmnd = args[0]
        result = []
        cmnds = main.find(cmnd)
        if not cmnds:
            try: cmnds = main[cmnd]
            except KeyError: cmnds = []
        return cmnds

            
## TESTBOT

class TestBot(Bot):

    """ USE WITH Bot.put() and Bot.read(). """

    testing = True
    cc = "!"
    default = ""

    def _raw(zelf, *args, **kwargs): print(args[0])

    def say(zelf, *args, **kwargs): logging.warn(args[1])

    def event(zelf, *args, **kwargs):
        o = Object()
        o.origin = "test@rechter"
        o._target = zelf
        o.loglevel = "warning"
        o.txt = args[0]
        o.parsed = txt_parse(zelf, o.txt)
        return o

## CMNDS

def join(event):
    if not event.parsed.args: return
    channel = event.parsed.args[0]
    for bot in fleet:
        if "join" not in bot: continue
        bot.join(channel)
        event.reply("join %s (%s)" % (channel, bot.server))

main.register("bot.join", join)

def part(event):
    if not event.parsed.args: return
    for bot in fleet:
        try: bot.part(event.parsed.args[0])
        except: continue
        event.reply("part %s (%s)" % (channel, bot.server))

main.register("bot.part", part)

def nick(event):
    if not event.parsed.args: return
    for bot in fleet:
        try: bot.nick(event.parsed.args[0])
        except: continue
        event.reply("nick %s" % bot.server)

main.register("bot.nick", nick)

def quit(event):
    if not event.parsed.args: return
    for bot in fleet:
        try: bot.quit(event.parsed.rest)
        except: continue
        event.reply("quit %s" % bot.server)
    
main.register("bot.quit", quit)
