# rechter/bots/xmpp.py
#
#

""" basic package for the program. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from rechter.cfg import cfg as maincfg
from rechter.main import main, fleet
from rechter.object import Object
from rechter.bots import Bot

import threading
import logging
import getpass
import queue

## XMPP

class XMPP(Bot):

    """ XMPP Bot. """

    cc = "none"
    default = ""

    def __init__(zelf, password, *args, **kwargs):
        import sleekxmpp
        from sleekxmpp import clientxmpp
        Bot.__init__(zelf, *args, **kwargs)
        if "port" not in zelf: zelf.port = maincfg.xmpp.port
        zelf._queue = queue.Queue()
        zelf.client = clientxmpp.ClientXMPP(zelf.user, password)
        zelf.add_handler("session_start", zelf.session_start)
        zelf.add_handler("message", zelf.messaged)
        zelf.add_handler('disconnected', zelf.disconnected)
        zelf.add_handler('connected', zelf.connected)
        zelf.add_handler('presence_available', zelf.presenced)
        zelf.add_handler('presence_dnd', zelf.presenced)
        zelf.add_handler('presence_xa', zelf.presenced)
        zelf.add_handler('presence_chat', zelf.presenced)
        zelf.add_handler('presence_away', zelf.presenced)
        zelf.add_handler('presence_unavailable', zelf.presenced)
        zelf.add_handler('presence_subscribe', zelf.presenced)
        zelf.add_handler('presence_subscribed', zelf.presenced)
        zelf.add_handler('presence_unsubscribe', zelf.presenced)
        zelf.add_handler('presence_unsubscribed', zelf.presenced)
        zelf.add_handler('groupchat_message', zelf.messaged)
        zelf.add_handler('groupchat_presence', zelf.presenced)
        zelf.add_handler('groupchat_subject', zelf.presenced)
        zelf.add_handler('failed_auth', zelf.failedauth)
        zelf.client.exception = zelf.exception
        zelf.client.use_signals()
        zelf._connected = threading.Event()
        zelf.channels = []

    def add_handler(zelf, key, value):
        logging.info("# %s/add %s" % (zelf.type, key))
        zelf.client.add_event_handler(key, value)

    def _raw(zelf, *args, **kwargs):
        logging.debug("> %s/write %s" % (zelf.type, args[0]))
        zelf.client.send_raw(args[0])

    def start(zelf, *args, **kwargs):
        try: zelf.connect()
        except: raise
        zelf.put(Bot.start, zelf)
        zelf.loop()

    def connect(zelf):
        logging.warn("# %s/connect %s" % (zelf.type, zelf.user))
        if maincfg.main.openfire:
            logging.warn("# %s/openfire" % zelf.type)
            import ssl
            zelf.client.ssl_version = ssl.PROTOCOL_SSLv3
            zelf.client.connect((zelf.server, zelf.port), use_ssl=True)
        else: zelf.client.connect((zelf.server, zelf.port), reattempt=True)

    def session_start(zelf, event):
        zelf.client.send_presence()
        zelf.ready()
        logging.warn("# %s/session %s" % (zelf.type, zelf.user))
         
    def exception(zelf, ex):
        zelf._error = ex
        logging.error("^ %s/error %s"% (zelf.type, str(ex)))

    def announce(zelf, *args, **kwargs):
        for channel in zelf.channels: zelf.say(channel, args[0])

    def failedauth(zelf, error, *args, **kwargs): zelf._error = error ; logging.error("^ %s/auth %s" % (zelf.type, str(error)))

    def failure(zelf, error, *args, **kwargs): zelf._error = error ; logging.error("^ %s/error %s" % (zelf.type, str(error)))

    def disconnected(zelf, *args, **kwargs):
        zelf._connected.clear()
        zelf._status = "disconnect"
        logging.warn("^ %s/disconnect %s" % (zelf.type, zelf.user))

    def connected(zelf, *args, **kwargs): 
        zelf._status = "connected"
        zelf._connected.set()
        logging.warn("# %s/connected %s" % (zelf.type, zelf.user))

    def loop(zelf, *args, **kwargs): zelf.client.process()

    def say(zelf, *args, **kwargs):
        to = args[0]
        txt = str(args[1])
        zelf.client.send_message(to, txt)
        logging.info("> %s/say %s" % (zelf.type, to))

    def event(zelf, *args, **kwargs): return zelf._queue.get()

    def messaged(zelf, data, *args, **kwargs):
        logging.debug("< %s/read %s" % (zelf.type, str(data["data"])))
        m = Object(**data)
        if m.type == "error": logging.error("^ %s" % m.error) ; return
        m.cc = zelf.cc
        m["from"] = str(m["from"])
        if zelf.user in m["from"]: logging.info("%s/%s %s" % (zelf.type, m.type, m["from"])) ; return
        m.origin = m["from"]
        m.channel = m.origin
        m.to = m.origin
        m.element = "message"
        m.txt = m["body"]
        m._target = zelf
        if '<delay xmlns="urn:xmpp:delay"' in str(data):
            logging.info("# %s/ignore %s %s" % (zelf.type, m.type, m.origin))
            return
        logging.info("< %s/%s %s" % (zelf.type, m.type, m.origin))
        zelf._queue.put_nowait(m)

    def presenced(zelf, data, *args, **kwargs):
        logging.info("< %s/%s %s" % (zelf.type, "presence", data["data"]))
        o = Object(data)
        o["from"] = str(o["from"])
        o.origin = o["from"]
        if "txt" not in o: o.txt = ""
        o.element = "presence"
        if o.type == 'subscribe':
            pres = Object({'to': o["from"], 'type': 'subscribed'})
            zelf.client.send_presence(pres)
            pres = Object({'to': o["from"], 'type': 'subscribe'})
            zelf.client.send_presence(pres)
        elif o.type == "unavailable":
            if o.origin in zelf.channels: zelf.channels.remove(o.origin)
        else:
            if o.origin != zelf.user and o.origin not in zelf.channels: zelf.channels.append(o.origin)
        o.no_dispatch = True
        o._target = zelf
        logging.info("< %s/%s %s" % (zelf.type, o.type, o.origin))
        zelf._queue.put_nowait(o)

    def exit(zelf, *args, **kwargs):
        zelf.client.disconnect()
        Bot.exit(zelf)
        zelf._queue.put(None)
        
def init(*args, **kwargs):
    cfg = Object().last("cfg", "xmpp")
    if not cfg: cfg = maincfg.xmpp ; cfg.save()
    password = getpass.getpass()
    xmpp = XMPP(password, **cfg)
    fleet.append(xmpp)
    main.put(xmpp.start)
