# rechter/dispatcher.py
#
#

""" basic package for the program. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from rechter.errors import NoFunction, NoEvent, RemoteDisconnect
from rechter.utils.name import naming, named, typed, naming
from rechter.utils.trace import get_plugname, get_strace
from rechter.utils.parse import txt_parse
from rechter.utils.time import elapsed
from rechter.errors import error

from rechter.object import Object

import collections
import threading
import logging
import random
import queue
import time

## TASKS

class Task(threading.Thread):

    """ the working unit of RECHTER.  """

    def __init__(zelf, *args, **kwargs):
        threading.Thread.__init__(zelf, *args, **kwargs)
        zelf._queue = queue.Queue()
        zelf.setDaemon(True)
        zelf.result = None
        
    def __iter__(zelf):
        for x in dir(zelf): yield x

    ## INPUT

    def put(zelf, *args, **kwargs):
        """ put arguments/kwargs to the Task. """
        zelf._last = time.time()
        zelf._queue.put((args, kwargs))
        return zelf

    def join(zelf, *args, **kwargs):
        threading.Thread.join(zelf)
        return zelf.result

    ## RUNNING

    def run(zelf, *args, **kwargs):
        """ loop that executes the (function, ding) pairs. """
        zelf.type = "Task"
        zelf._start = time.time()
        zelf._last = time.time()
        zelf._status = "ok"
        zelf.name = ""
        while zelf._status:
            arglist, kwargdict = zelf._queue.get()
            if not arglist or not arglist[0]: break
            zelf.func, *arguments = arglist
            zelf.name = named(zelf.func)
            zelf.setName(zelf.name)
            zelf.result = zelf.func(*arguments, **kwargdict)
            #try: zelf.result = zelf.func(*arguments, **kwargdict)
            #except Exception as ex: zelf._status = ex ; error("", zelf.name) ; break
            if arguments: logging.debug(arguments[0])
            if zelf._queue.empty(): break
        logging.info("! Task/run %s" % zelf.name)
        
    def exit(zelf, *args, **kwargs):
        zelf._status = ""
        zelf.put((None, {}))

## DISPATCHER

class Dispatcher(Object):

    """ the dispatcher delegates the workload to the Runners. Runner gets instantiated when needed. """

    cc = "!"
    default = ""

    def __init__(zelf, *args, **kwargs):
        Object.__init__(zelf, *args, **kwargs)
        zelf._input = queue.Queue()
        zelf._output = queue.Queue()
        zelf._ready = threading.Event()

    ## INPUT

    def put(zelf, *args, **kwargs):
        task = Task()
        task.start()
        task.put(*args, **kwargs)
        return task

    ## RESULTS

    def collect(zelf, *args, **kwargs):
        thrs = args[0]
        if not thrs: return []
        result = []
        for thr in thrs:
            if thr:
                res = thr.join()
                if res: result.append(res)
        return result

    ## STOP

    def stop(zelf, *args, **kwargs):
        zelf._status = ""
        zelf.put(None)

## SHUTDOWN

def shutdown(*args, **kwargs):
    if args: name = args[0]
    else: name = ""
    for thr in threading.enumerate():
        if str(thr).startswith("<_"): continue
        if name and name not in str(thr): continue
        if "stop" in dir(thr): thr.stop()
        elif "exit" in dir(thr): thr.exit()
        elif "cancel" in dir(thr): thr.cancel()
        else: thr.join(0.1)
        logging.warn("# stop %s" % naming(thr))
    return True

