# main.py
#
#

""" main module. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from rechter.utils.opts import make_opts, opts_defs
from rechter.utils.other import hello, show_eggs
from rechter.defines import GREEN, ENDC
from rechter.dispatcher import shutdown
from rechter.object import Object
from rechter.plugins import Plugins
from rechter.log import log_init
from rechter.cfg import cfg

import logging
import time
import sys
import os

## PATH

def rechterpath():
    for path in sys.path:
        if "rechter" in path and "egg" in path: yield path  

## MAIN 

class Main(Plugins):

    def boot(zelf, *args, **kwargs):
        """
             Enable runtime services of the RECHTER bot.

             >>> main = Main()
             >>> main.boot("RECHTER")
             >>> main.wait()

        """ 
        zelf._start = time.time()
        opts, arglist = make_opts(opts_defs)
        cfg.main.update(vars(opts))
        for key,value in kwargs.items():
            if not getattr(cfg.main , key, ""): cfg.main[key] = value
        if not cfg.main.workdir: cfg.main.workdir = os.path.expanduser("~/.rechter")
        cfg.main.runargs = arglist
        log_init(cfg.main.loglevel, colors=cfg.main.colors)
        if cfg.main.shell: hello(*args, **cfg.main)
        if cfg.main.path:
            for path in sys.path: print(path)
            print()
        plugins.boot(cfg.main.packages)
        if "all" in cfg.main.modules: plugins.loaddir(os.getcwd())
        else:
            for mod in cfg.main.modules.split(","):
                if not mod: continue
                try: plugins.load(mod)
                except Exception as ex: print(ex)
        plugins.init("system,clock")
        zelf.ready()

    def shutdown(zelf, *args, **kwargs):
        logging.warn("# shutdown")
        plugins.shutdown()
        shutdown()
        zelf.stop()
        time.sleep(0.01)

## DEFINES

main = Main()
plugins = Plugins()
fleet = []
