# plugins.py
#
#

""" rechter plugins. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from rechter.dispatcher import Dispatcher
from rechter.utils.name import named
from rechter.object import Object
from rechter.main import shutdown
from rechter.errors import error
from rechter.cfg import cfg

import importlib
import logging
import pkgutil
import os

## PLUGINS

class Plugins(Dispatcher):

    """ register plugins. """

    plugnames = ""

    ## INIT

    def init(zelf, *args, **kwargs):
        """ initialize plugins, """
        plugnames = args[0]
        modnames = plugnames.split(",") 
        for modname in sorted(modnames, reverse=True):
            if not modname: continue
            objs = zelf.find(modname)
            for obj in objs:
                logging.warn("# %s/init %s" % (zelf.type, modname))
                if modname == "xmpp": obj.init() ; continue
                init = getattr(obj, "init", None)
                if init: init()
        zelf.ready()

    def boot(zelf, *args, **kwargs):
        from rechter.cfg import cfg
        packages = args[0]
        for package in packages:
            if package in cfg.main.exclude and package.split(",")[-1] not in cfg.main.modules:
                logging.warn("# Plugins/ignore %s" % package)
                continue
            mod = zelf.load(package)
            register = getattr(mod, "register", None)
            if register:
                logging.warn("# %s/register %s" % (zelf.type, package))
                register()
            init = getattr(mod, "init", None)
            if init:
                logging.warn("# %s/init %s" % (zelf.type, package))
                init()

    def shutdown(zelf, *args, **kwargs):
        for name, obj in zelf.items():
            if "shutdown" in dir(obj):
                logging.warn("# %s/shutdown" % zelf.type)
                obj.shutdown()

    ## SEARCH

    def search(zelf, *args, **kwargs):
        """ find matching name. """
        n = args[0]
        for name in zelf.names():
            if n in name: yield name

    ## LOAD

    def load(zelf, *args, **kwargs):
        """ load plugins from the plugin directory. """
        from rechter.cfg import cfg
        modname = args[0]
        logging.warn("# %s/load %s" % (zelf.type, modname))
        zelf[modname] = mod = zelf.loading(modname)
        if "__path__" in dir(mod):
            for pkg in pkgutil.walk_packages(mod.__path__, modname + "."):
                name = pkg[1]
                zelf[name] = zelf.loading(name)
        return zelf[modname]

    def loading(zelf, name, package=None, force=True):
        """ load module. """
        logging.debug("! %s/load %s" % (zelf.type, name))
        if name in zelf and not force: mod = zelf.reload(name)
        else: mod = importlib.import_module(name, package)
        return mod

    def loaddir(zelf, *args, **kwargs):
        dirname = args[0]
        for fn in os.listdir(dirname):
            if not fn.endswith(".egg"): continue
            name = fn.split("-")[0]
            try: zelf[name] = zelf.loading(name)
            except Exception as ex: print(ex)

    ## RELOAD

    def reload(zelf, *args, **kwargs):
        name = args[0]
        logging.warn("# %s/reload %s" % (zelf.type, name))
        zelf[name] = importlib.reload(zelf[name])
        return zelf[name]

    def reloading(zelf, *args, **kwargs):
        names = zelf.find(args[0])
        for name in names: zelf.reload(name)

## CMNDS

def plugins_stop(event):
    from rechter.main import plugins
    name = event.parsed.rest
    if name and plugins.search(name): shutdown(name) ; event.ok(name)
    logging.warn("# Plugins/stop %s" % name)

def plugins_start(event):
    from rechter.main import plugins
    args = event.parsed.args
    if args:
        pluginnames = plugins.search(*args)
        for pname in pluginnames:
            plugins.load(pname)
            plugins.init(event.parsed.args[0])
            event.ok(pname)
    logging.warn("# Plugins/start %s" % ", ".join(pluginnames))

def plugins_reload(event):
    from rechter.main import plugins
    args = event.parsed.args
    if args:
        result = plugins.reloading(*args)
        if result: event.ok(named(result))
    logging.warn("# Plugins/reload")

def plugins_load(event):
    from rechter.main import plugins
    name = event.parsed.rest
    if name:
        try:
            result = plugins.load(name)
            if result: event.ok(named(result))
        except: error(name) 
    logging.warn("# Plugins/load")

## REGISTER

def register(*args, **kwargs):
    from rechter.main import main
    main.register("plugins.reload", plugins_reload)
    main.register("plugins.load", plugins_load)
    main.register("plugins.start", plugins_start)
    main.register("plugins.stop", plugins_stop)
