# rechter/services/http.py
#
#

""" RECHTER INTERFACE. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from rechter.utils.trace import get_exception

from rechter.main import main
from rechter.object import Object
from rechter import __version__
from rechter.cfg import cfg

import logging
import time

from http.server import HTTPServer, BaseHTTPRequestHandler

## DEFINES

port = 10102
server = None

## APIServer class

class REST(HTTPServer, Object):

    """ REST server """

    allow_reuse_address = True
    daemon_thread = True

    def __init__(zelf, *args, **kwargs):
        """ start the HTTP server. """
        HTTPServer.__init__(zelf, *args, **kwargs)
        Object.__init__(zelf)
        zelf.host = args[0]
        zelf._status = "start"
        zelf._last = time.time()
        zelf._start = time.time()

    def exit(zelf):
        """ shutdown. """
        logging.warn("# REST/end")
        zelf._status = ""
        time.sleep(0.2)
        zelf.shutdown()

    def start(zelf, *args, **kwargs): 
        """ serving requests. """
        logging.warn("# REST/start http://%s:%s" % zelf.host)
        zelf._status = "ok"
        zelf.ready()
        zelf.serve_forever()

    def request(zelf):
        zelf._last = time.time()

    def error(zelf, request, addr):
        """ log the error """
        ex = get_exception()
        logging.warn('# REST/error %s error %s' % (addr, ex))

## RestReqeustHandler class

class RESTHandler(BaseHTTPRequestHandler):

    """ timeserver request handler class. """

    def setup(zelf):
        """ called on each incoming request. """
        BaseHTTPRequestHandler.setup(zelf)
        zelf._ip = zelf.client_address[0]
        zelf._size = 0

    def write_header(zelf, type='text/plain'):
        """ write headers to the client. """
        zelf.send_response(200)
        zelf.send_header('Content-type', '%s; charset=%s ' % (type, "utf-8"))
        zelf.send_header('Server', __version__)
        zelf.end_headers()

    def do_GET(zelf, *args):
        """ handle a REST request. """
        root = main.root()
        filename = ""
        for fn in main.all():
            if fn.endswith(zelf.path): filename = fn ; break
        try: f = open(filename, "r") ; txt = f.read() ; f.close()
        except (TypeError, FileNotFoundError): zelf.send_response(404) ; zelf.end_headers() ; return
        txt = txt.replace("\\n", "\n")
        txt = txt.replace("\\t", "\t")
        zelf.write_header()
        zelf.wfile.write(bytes(txt, "utf-8"))
        zelf.wfile.flush()

    def log(zelf, code):
        """ log the request. """
        try: ua = zelf.headers['user-agent']
        except: ua = "-"
        try: rf = zelf.headers['referer']
        except: rf = "-"
        logging.warn('# REST/log %s code %s path %s' % (zelf.address_string(), code, zelf.path))


def init(*args, **kwargs):
    """ initialize the HTTP server. """
    server = REST((cfg.rest.hostname, int(cfg.main.port or cfg.rest.port)), RESTHandler)
    main.put(server.start)
