# rechter/services/spider.py
#
#

""" Spider plugin. """

## IMPORTS

from rechter.utils.url import parse_url, need_redirect, get_url
from rechter.errors import error
from rechter.main import main
from rechter.object import Object

import logging
import time

## DEFINES

url_list = []

## UTILS

def parse_urls(*args, **kwargs): 
    import bs4
    url = args[0]
    try: content = args[1]
    except: content = get_url("GET", url).read()
    basepath, base, root, file = parse_url(url)
    s = bs4.BeautifulSoup(content)
    urls = []
    tags = s('a')
    for tag in tags:
        href = tag.get("href")
        if href: 
            href = href.split("#")[0]
            if not href: continue
            if ".." in href: continue
            if href.startswith("mailto"): continue
            if not "http" in href:
                if href.startswith("/"): href = root + href
                else: href = base + "/" + href
                if not root in href: continue
            if href not in urls: urls.append(href)
    logging.warn("# Utils/parseurls %s" % len(urls))
    return urls

## SPIDER

class Spider(Object):

    def __init__(zelf, *args, **kwargs):
        Object.__init__(zelf, *args, **kwargs)
        zelf.errors = []
        zelf.urls = []
        zelf.url = Object()
        zelf.followed = []
        zelf.speed = 0
        zelf.depth = 10 

    def boot(zelf, *args, **kwargs):
        for fn in main.all("spider"):
            logging.warn("spider %s" % fn)
            obj = Object().load(fn)
            zelf.crawl(obj.url)

    def crawl(zelf, *args, **kwargs):
        time.sleep(zelf.speed)
        zelf.speed += 1.0
        try: url, search = args
        except ValueError: url = args[0] ; search = ""
        logging.warn("crawl %s" % url)
        urls = []
        if not zelf.url:
           zelf.url.url = url
           zelf.url.basepath, zelf.url.base, zelf.url.root, zelf.url.file = parse_url(zelf.url.url)
        pnr = len(url.split("/"))
        if pnr >  zelf.depth: logging.warn("%s max depth " % url) ; return
        if url not in zelf.urls and url not in url_list: 
            zelf.urls.append(url)
            try: content = get_url("GET", url)
            except: error() ; return
            newurl = need_redirect(content)
            if newurl:
                logging.warn("redirecting to %s" % newurl)
                content = get_url("GET", newurl)
            newurl2 = need_redirect(content)
            if newurl2:
                logging.warn("redirecting to %s" % newurl2)
                content = get_url("GET", newurl2)
            data = content.read()
            urls = parse_urls(url, data)
            for u in urls:
                if u in zelf.urls: continue
                if not zelf.url.base in u: continue
                if u in zelf.errors: continue
                zelf.crawl(u, search)
        o = Object()    
        o.prefix = "spider.data"
        o.services = "spider"
        o.urls = urls
        o.sync()
        return urls

## CMNDS

def spider(event):
    try: url = event.parsed.args[0]
    except IndexError: return
    if "http" not in url: return
    try: search = event.parsed.args[1]
    except IndexError: search = ""
    urls = main.all("spider")
    if url not in urls:
        object = Object()
        object.prefix = "spider"
        object.url = url
        object.search = search
        object.save()
    s = Spider()
    s.crawl(url, search)
    for url in s.urls: event.reply(url)

main.register("spider.spider", spider)

## INIT

def init(*args, **kwargs):
    spider = Spider()
    main.put(spider.boot)
